"""
The module that provides generating methods for all parts
of the pdelab driver.

Currently, these are hardcoded as strings. It would be possible
to switch these to cgen expression. OTOH, there is not much to be
gained there.
"""
from dune.perftool.generation import (generator_factory,
                                      include_file,
                                      symbol,
                                      preamble,
                                      )
from dune.perftool.options import get_option


def has_constraints(element):
    from ufl import MixedElement, VectorElement
    if isinstance(element, MixedElement) and not isinstance(element, VectorElement):
        return any(has_constraints(el) for el in element.sub_elements())
    return hasattr(element, "dirichlet_constraints")


# Those symbol generators that depend on the meta data we hacked into
# the ufl.FiniteElement should use fem_metadata_dependent_symbol instead of symbol.
def get_constraints_metadata(element):
    """
    Normally, we suppress the constraints information on a finite element.
    However, we might want to use it explicitly in those case where it matters
    This function does the job!
    """
    dc = getattr(element, 'dirichlet_constraints', None)
    de = getattr(element, 'dirichlet_expression', None)
    return (element, (dc, de))


fem_metadata_dependent_symbol = generator_factory(item_tags=("symbol",),
                                                  cache_key_generator=get_constraints_metadata,
                                                  )


# Have a global variable with the entire form data. This allows functions that depend
# deterministically on the entire data set to directly access it instead of passing it
# through the entire generator chain.
_form = None


# Have a function access this global data structure
def set_form(f):
    global _form
    _form = f


def is_linear(form):
    '''Test if form is linear in test function'''
    from ufl import derivative
    from ufl.algorithms import expand_derivatives
    jacform = expand_derivatives(derivative(form, form.coefficients()[0]))
    for coeff in jacform.coefficients():
        if 0 == coeff.count():
            return False

    return True


def isLagrange(fem):
    return fem._short_name is 'CG'


def isSimplical(fem):
    return any(fem._cell._cellname in x for x in ["triangle", "tetrahedron"])


def isQuadrilateral(fem):
    return any(fem._cell._cellname in x for x in ["vertex", "interval", "quadrilateral", "hexahedron"])


def isPk(fem):
    return isLagrange(fem) and isSimplical(fem)


def isQk(fem):
    return isLagrange(fem) and isQuadrilateral(fem)


def isDG(fem):
    return fem._short_name is 'DG'


def FEM_name_mangling(fem):
    from ufl import MixedElement, VectorElement, FiniteElement, TensorElement
    if isinstance(fem, VectorElement):
        return FEM_name_mangling(fem.sub_elements()[0]) + "_" + str(fem.num_sub_elements())
    if isinstance(fem, TensorElement):
        return FEM_name_mangling(fem.sub_elements()[0]) + "_" + "_".join(str(i) for i in fem.value_shape())
    if isinstance(fem, MixedElement):
        name = ""
        for elem in fem.sub_elements():
            if name is not "":
                name = name + "_"
            name = name + FEM_name_mangling(elem)
        return name
    if isinstance(fem, FiniteElement):
        if isPk(fem):
            return "P" + str(fem._degree)
        if isQk(fem):
            return "Q" + str(fem._degree)
        if isDG(fem):
            return "DG" + str(fem._degree)

    raise NotImplementedError("FEM NAME MANGLING")


@symbol
def name_inifile():
    # TODO pass some other option here.
    return "argv[1]"


@preamble
def parse_initree(varname):
    include_file("dune/common/parametertree.hh", filetag="driver")
    include_file("dune/common/parametertreeparser.hh", filetag="driver")
    filename = name_inifile()
    return ["Dune::ParameterTree initree;", "Dune::ParameterTreeParser::readINITree({}, {});".format(filename, varname)]


@symbol
def name_initree():
    parse_initree("initree")
    # TODO we can get some other ini file here.
    return "initree"


@preamble
def define_dimension(name):
    return "static const int {} = {};".format(name, _form.ufl_cell().geometric_dimension())


@symbol
def name_dimension():
    define_dimension("dim")
    return "dim"


@preamble
def typedef_grid(name):
    dim = name_dimension()
    if any(_form.ufl_cell().cellname() in x for x in ["vertex", "interval", "quadrilateral", "hexahedron"]):
        gridt = "Dune::YaspGrid<{}>".format(dim)
        include_file("dune/grid/yaspgrid.hh", filetag="driver")
    else:
        if any(_form.ufl_cell().cellname() in x for x in ["triangle", "tetrahedron"]):
            # gridt = "Dune::UGGrid<{}>".format(dim)
            # include_file("dune/grid/uggrid.hh", filetag="driver")
            gridt = "Dune::ALUGrid<{}, {}, Dune::simplex, Dune::conforming>".format(dim, dim)
            include_file("dune/alugrid/grid.hh", filetag="driver")
        else:
            raise ValueError("Cant match your geometry with a DUNE grid. Please report bug.")
    return "typedef {} {};".format(gridt, name)


@symbol
def type_grid():
    typedef_grid("Grid")
    return "Grid"


@preamble
def define_grid(name):
    include_file("dune/testtools/gridconstruction.hh", filetag="driver")
    ini = name_initree()
    _type = type_grid()
    return ["IniGridFactory<{}> factory({});".format(_type, ini),
            "std::shared_ptr<{}> grid = factory.getGrid();".format(_type)]


@symbol
def name_grid():
    define_grid("grid")
    return "grid"


@preamble
def typedef_leafview(name):
    grid = type_grid()
    return "typedef {}::LeafGridView {};".format(grid, name)


@symbol
def type_leafview():
    typedef_leafview("GV")
    return "GV"


@preamble
def define_leafview(name):
    _type = type_leafview()
    grid = name_grid()
    return "{} {} = {}->leafGridView();".format(_type, name, grid)


@symbol
def name_leafview():
    define_leafview("gv")
    return "gv"


@preamble
def typedef_vtkwriter(name):
    include_file("dune/grid/io/file/vtk/subsamplingvtkwriter.hh", filetag="driver")
    gv = type_leafview()
    return "typedef Dune::SubsamplingVTKWriter<{}> {};".format(gv, name)


@symbol
def type_vtkwriter():
    typedef_vtkwriter("VTKWriter")
    return "VTKWriter"


@preamble
def define_subsamplinglevel(name):
    ini = name_initree()
    return "int {} = {}.get<int>(\"vtk.subsamplinglevel\", 0);".format(name, ini)


@symbol
def name_subsamplinglevel():
    define_subsamplinglevel("sublevel")
    return "sublevel"


@preamble
def define_vtkwriter(name):
    _type = type_vtkwriter()
    gv = name_leafview()
    subsamp = name_subsamplinglevel()
    return "{} {}({}, {});".format(_type, name, gv, subsamp)


@symbol
def name_vtkwriter():
    define_vtkwriter("vtkwriter")
    return "vtkwriter"


@preamble
def typedef_domainfield(name):
    gridt = type_grid()
    return "typedef {}::ctype {};".format(gridt, name)


@symbol
def type_domainfield():
    typedef_domainfield("DF")
    return "DF"


@preamble
def typedef_range(name):
    return "typedef double {};".format(name)


@symbol
def type_range():
    typedef_range("R")
    return "R"


@preamble
def typedef_fem(expr, name):
    gv = type_leafview()
    df = type_domainfield()
    r = type_range()
    dim = name_dimension()
    if isPk(expr):
        include_file("dune/pdelab/finiteelementmap/pkfem.hh", filetag="driver")
        return "typedef Dune::PDELab::PkLocalFiniteElementMap<{}, {}, {}, {}> {};".format(gv, df, r, expr._degree, name)
    if isQk(expr):
        include_file("dune/pdelab/finiteelementmap/qkfem.hh", filetag="driver")
        return "typedef Dune::PDELab::QkLocalFiniteElementMap<{}, {}, {}, {}> {};".format(gv, df, r, expr._degree, name)
    if isDG(expr):
        if isQuadrilateral(expr):
            include_file("dune/pdelab/finiteelementmap/qkdg.hh", filetag="driver")
            # TODO allow switching the basis here!
            return "typedef Dune::PDELab::QkDGLocalFiniteElementMap<{}, {}, {}, {}> {};".format(df, r, expr._degree, dim, name)
        if isSimplical(expr):
            include_file("dune/pdelab/finiteelementmap/opbfem.hh", filetag="driver")
            return "typedef Dune::PDELab::OPBLocalFiniteElementMap<{}, {}, {}, {}, Dune::GeometryType::simplex> {};".format(df, r, expr._degree, dim, name)
        raise NotImplementedError("Geometry type not known in code generation")

    raise NotImplementedError("FEM not implemented in dune-perftool")


@symbol
def type_fem(expr):
    name = "{}_FEM".format(FEM_name_mangling(expr).upper())
    typedef_fem(expr, name)
    return name


@preamble
def define_fem(expr, name):
    femtype = type_fem(expr)
    if isDG(expr):
        return "{} {};".format(femtype, name)
    else:
        gv = name_leafview()
        return "{} {}({});".format(femtype, name, gv)


@symbol
def name_fem(expr):
    name = "{}_fem".format(FEM_name_mangling(expr).lower())
    define_fem(expr, name)
    return name


@preamble
def typedef_vectorbackend(name):
    include_file("dune/pdelab/backend/istl.hh", filetag="driver")
    return "typedef Dune::PDELab::istl::VectorBackend<Dune::PDELab::istl::Blocking::none, 1> {};".format(name)


@symbol
def type_vectorbackend():
    typedef_vectorbackend("VectorBackend")
    return "VectorBackend"


@symbol
def type_orderingtag():
    return "Dune::PDELab::LexicographicOrderingTag"


@preamble
def typedef_dirichlet_constraintsassembler(name):
    include_file("dune/pdelab/constraints/conforming.hh", filetag="driver")
    return "typedef Dune::PDELab::ConformingDirichletConstraints {};".format(name)


@preamble
def typedef_no_constraintsassembler(name):
    return "typedef Dune::PDELab::NoConstraints {};".format(name)


@symbol
def type_constraintsassembler(dirichlet):
    if dirichlet[1]:
        typedef_dirichlet_constraintsassembler("DirichletConstraintsAssember")
        return "DirichletConstraintsAssember"
    else:
        typedef_no_constraintsassembler("NoConstraintsAssembler")
        return "NoConstraintsAssembler"


@preamble
def typedef_constraintscontainer(element, name):
    gfs = type_gfs(element)
    r = type_range()
    return "typedef {}::ConstraintsContainer<{}>::Type {};".format(gfs, r, name)


@symbol
def type_constraintscontainer(element):
    name = "{}_cc".format(FEM_name_mangling(element)).upper()
    typedef_constraintscontainer(element, name)
    return name


@preamble
def define_constraintscontainer(expr, name):
    cctype = type_constraintscontainer(expr)
    return ["{} {};".format(cctype, name), "{}.clear();".format(name)]


# TODO this does not yet fully support having same elements with different constraints!
@fem_metadata_dependent_symbol
def name_constraintscontainer(expr):
    name = "{}{}_cc".format(FEM_name_mangling(expr), "_dirichlet" if has_constraints(expr) else "").lower()
    define_constraintscontainer(expr, name)
    return name


@preamble
def define_intersection_lambda(expression, name):
    if expression is None:
        return "auto {} = [&](const auto& x){{ return 0; }};".format(name)
    if expression.is_global:
        return "auto {} = [&](const auto& x){{ {} }};".format(name, expression.c_expr)
    else:
        return "auto {} = [&](const auto& is, const auto& x){{ {} }};".format(name, expression.c_expr)


@symbol
def name_bctype_lambda(name, dirichlet):
    name = name + "_lambda"
    define_intersection_lambda(dirichlet, name)
    return name


@preamble
def define_bctype_function(dirichlet, name):
    gv = name_leafview()
    bctype_lambda = name_bctype_lambda(name, dirichlet)
    include_file('dune/pdelab/function/callableadapter.hh', filetag='driver')
    return "auto {} = Dune::PDELab::makeBoundaryConditionFromCallable({}, {});".format(name,
                                                                                       gv,
                                                                                       bctype_lambda,
                                                                                       )


@preamble
def define_powergfs_constraints(name, child, dim):
    include_file('dune/pdelab/constraints/common/constraintsparameters.hh', filetag='driver')
    return "Dune::PDELab::PowerConstraintsParameters<decltype({}), {}> {}({});".format(child, dim, name, child)


@preamble
def define_compositegfs_constraints(name, *children):
    include_file('dune/pdelab/constraints/common/constraintsparameters.hh', filetag='driver')
    return "Dune::PDELab::CompositeConstraintsParameters<{}> {}({});".format(', '.join('decltype({})'.format(c) for c in children),
                                                                             name,
                                                                             ', '.join(c for c in children)
                                                                             )


@symbol
def name_bctype_function(expr):
    from ufl import MixedElement, VectorElement, TensorElement
    if isinstance(expr, (VectorElement, TensorElement)):
        element, (dirichlet, _) = get_constraints_metadata(expr)
        # get the correct name from the corresponding UFL file
        from dune.perftool.generation import get_global_context_value
        name = get_global_context_value("data").object_names.get(id(dirichlet), "everywhere")
        define_bctype_function(dirichlet, name)
        pgfs_name = '{}_{}'.format(name, expr.num_sub_elements())
        define_powergfs_constraints(pgfs_name, name, expr.num_sub_elements())
        return pgfs_name
    if isinstance(expr, MixedElement):
        children = tuple(name_bctype_function(el) for el in expr.sub_elements())
        name = '_'.join(children)
        define_compositegfs_constraints(name, *children)
        return name

    # So, this seems to be a leaf!
    element, (dirichlet, _) = get_constraints_metadata(expr)

    # get the correct name from the corresponding UFL file
    from dune.perftool.generation import get_global_context_value
    name = get_global_context_value("data").object_names.get(id(dirichlet), "everywhere")

    define_bctype_function(dirichlet, name)
    return name


@preamble
def assemble_constraints(name, expr):
    bctype_function = name_bctype_function(expr)
    gfs = name_gfs(expr)
    return "Dune::PDELab::constraints({}, {}, {});".format(bctype_function,
                                                           gfs,
                                                           name,
                                                           )


@symbol
def name_assembled_constraints(expr):
    name = name_constraintscontainer(expr)
    if has_constraints(expr):
        assemble_constraints(name, expr)
    return name


@preamble
def typedef_gfs(element, dirichlet, name):
    vb = type_vectorbackend()
    from ufl import FiniteElement, MixedElement, VectorElement, EnrichedElement, RestrictedElement, TensorElement
    if isinstance(element, FiniteElement):
        gv = type_leafview()
        fem = type_fem(element)
        cass = type_constraintsassembler(dirichlet)
        return "typedef Dune::PDELab::GridFunctionSpace<{}, {}, {}, {}> {};".format(gv, fem, cass, vb, name)
    if isinstance(element, (VectorElement, TensorElement)):
        include_file("dune/pdelab/gridfunctionspace/vectorgridfunctionspace.hh", filetag="driver")
        gv = type_leafview()
        fem = type_fem(element.sub_elements()[0])
        # TODO implement symmetry here
        dim = element.num_sub_elements()
        cass = type_constraintsassembler(dirichlet)
        return "typedef Dune::PDELab::VectorGridFunctionSpace<{}, {}, {}, {}, {}, {}> {};".format(gv, fem, dim, vb, vb, cass, name)
    if isinstance(element, MixedElement):
        ot = type_orderingtag()
        args = ", ".join(type_gfs(e) for e in element.sub_elements())
        return "typedef Dune::PDELab::CompositeGridFunctionSpace<{}, {}, {}> {};".format(vb, ot, args, name)
    if isinstance(element, EnrichedElement):
        raise NotImplementedError("Dune does not support enriched elements!")
    if isinstance(element, RestrictedElement):
        raise NotImplementedError("Dune does not support restricted elements!")


@fem_metadata_dependent_symbol
def type_gfs(expr):
    element, dirichlet = get_constraints_metadata(expr)
    name = "{}{}_gfs".format(FEM_name_mangling(element), "_dirichlet" if dirichlet else "").upper()
    typedef_gfs(element, dirichlet, name)
    return name


@preamble
def define_gfs(element, dirichlet, name):
    gfstype = type_gfs(element)
    from ufl import FiniteElement, MixedElement, VectorElement, EnrichedElement, RestrictedElement, TensorElement
    if isinstance(element, FiniteElement):
        gv = name_leafview()
        fem = name_fem(element)
        return "{} {}({}, {});".format(gfstype, name, gv, fem)
    if isinstance(element, (VectorElement, TensorElement)):
        gv = name_leafview()
        fem = name_fem(element.sub_elements()[0])
        return "{} {}({}, {});".format(gfstype, name, gv, fem)
    if isinstance(element, MixedElement):
        args = ", ".join(name_gfs(childgfs) for childgfs in element.sub_elements())
        return "{} {}({});".format(gfstype, name, args)
    raise NotImplementedError("Dune does not support this type of finite element!")


@fem_metadata_dependent_symbol
def name_gfs(expr):
    element, dirichlet = get_constraints_metadata(expr)
    name = "{}{}_gfs".format(FEM_name_mangling(element), "_dirichlet" if dirichlet else "").lower()
    define_gfs(element, dirichlet, name)
    return name


@preamble
def define_dofestimate(name):
    # Provide a worstcase estimate for the number of entries per row based on the given gridfunction space and cell geometry
    if isQuadrilateral(_form.coefficients()[0].ufl_element()):
        geo_factor = "4"
    else:
        geo_factor = "6"
    gfs = name_gfs(_form.coefficients()[0].ufl_element())
    ini = name_initree()
    return ["int generic_dof_estimate =  {} * {}.maxLocalSize();".format(geo_factor, gfs),
            "int {} = {}.get<int>(\"istl.number_of_nnz\", generic_dof_estimate);".format(name, ini)]


@symbol
def name_dofestimate():
    define_dofestimate("dofestimate")
    return "dofestimate"


@preamble
def typedef_matrixbackend(name):
    include_file("dune/pdelab/backend/istl.hh", filetag="driver")
    return "typedef Dune::PDELab::istl::BCRSMatrixBackend<> {};".format(name)


@symbol
def type_matrixbackend():
    typedef_matrixbackend("MatrixBackend")
    return "MatrixBackend"


@preamble
def define_matrixbackend(name):
    mbtype = type_matrixbackend()
    dof = name_dofestimate()
    return "{} {}({});".format(mbtype, name, dof)


@symbol
def name_matrixbackend():
    define_matrixbackend("mb")
    return "mb"


@symbol
def type_parameters():
    return "LocalOperatorParams"


@preamble
def define_parameters(name):
    partype = type_parameters()
    return "{} {};".format(partype, name)


@symbol
def name_parameters():
    define_parameters("params")
    return "params"


@preamble
def typedef_localoperator(name):
    # No Parameter class here, yet
    # params = type_parameters()
    # return "typedef LocalOperator<{}> {};".format(params, name)
    ugfs = type_gfs(_form.coefficients()[0].ufl_element())
    vgfs = type_gfs(_form.arguments()[0].ufl_element())
    from dune.perftool.options import get_option
    include_file(get_option('operator_file'), filetag="driver")
    return "using {} = LocalOperator<{}, {}>;".format(name, ugfs, vgfs)


@symbol
def type_localoperator():
    typedef_localoperator("LOP")
    return "LOP"


@preamble
def define_localoperator(name):
    loptype = type_localoperator()
    ini = name_initree()
    params = name_parameters()
    return "{} {}({}, {});".format(loptype, name, ini, params)


@symbol
def name_localoperator():
    define_localoperator("lop")
    return "lop"


@preamble
def typedef_gridoperator(name):
    ugfs = type_gfs(_form.coefficients()[0].ufl_element())
    vgfs = type_gfs(_form.arguments()[0].ufl_element())
    lop = type_localoperator()
    ucc = type_constraintscontainer(_form.coefficients()[0].ufl_element())
    vcc = type_constraintscontainer(_form.arguments()[0].ufl_element())
    mb = type_matrixbackend()
    df = type_domainfield()
    r = type_range()
    include_file("dune/pdelab/gridoperator/gridoperator.hh", filetag="driver")
    return "typedef Dune::PDELab::GridOperator<{}, {}, {}, {}, {}, {}, {}, {}, {}> {};".format(ugfs, vgfs, lop, mb, df, r, r, ucc, vcc, name)


@symbol
def type_gridoperator():
    typedef_gridoperator("GO")
    return "GO"


@preamble
def define_gridoperator(name):
    gotype = type_gridoperator()
    ugfs = name_gfs(_form.coefficients()[0].ufl_element())
    ucc = name_assembled_constraints(_form.coefficients()[0].ufl_element())
    vgfs = name_gfs(_form.arguments()[0].ufl_element())
    vcc = name_assembled_constraints(_form.arguments()[0].ufl_element())
    lop = name_localoperator()
    mb = name_matrixbackend()
    return ["{} {}({}, {}, {}, {}, {}, {});".format(gotype, name, ugfs, ucc, vgfs, vcc, lop, mb),
            "std::cout << \"gfs with \" << {}.size() << \" dofs generated  \"<< std::endl;".format(ugfs),
            "std::cout << \"cc with \" << {}.size() << \" dofs generated  \"<< std::endl;".format(ucc)]


@symbol
def name_gridoperator():
    define_gridoperator("go")
    return "go"


@preamble
def typedef_vector(name):
    gotype = type_gridoperator()
    return "typedef {}::Traits::Domain {};".format(gotype, name)


@symbol
def type_vector():
    typedef_vector("V")
    return "V"


@preamble
def define_vector(name):
    vtype = type_vector()
    gfs = name_gfs(_form.coefficients()[0].ufl_element())
    return ["{} {}({});".format(vtype, name, gfs), "{} = 0.0;".format(name)]


@preamble
def define_boundary_lambda(boundary, name):
    if boundary is None:
        return "auto {} = [&](const auto& x){{ return 0.0; }};".format(name)
    if boundary.is_global:
        return "auto {} = [&](const auto& x){{ {} }};".format(name, boundary.c_expr)
    else:
        return "auto {} = [&](const auto& e, const auto& x){{ {} }};".format(name, boundary.c_expr)


@symbol
def name_boundary_lambda(boundary, name):
    define_boundary_lambda(boundary, name + "lambda")
    return name + "lambda"


@preamble
def define_boundary_function(boundary, name):
    gv = name_leafview()
    lambdaname = name_boundary_lambda(boundary, name)
    include_file('dune/pdelab/function/callableadapter.hh', filetag='driver')
    return "auto {} = Dune::PDELab::makeGridFunctionFromCallable({}, {});".format(name,
                                                                                  gv,
                                                                                  lambdaname,
                                                                                  )


@preamble
def define_compositegfs_parameterfunction(name, *children):
    return "Dune::PDELab::CompositeGridFunction<{}> {}({});".format(', '.join('decltype({})'.format(c) for c in children),
                                                                    name,
                                                                    ', '.join(children))


@symbol
def name_boundary_function(expr):
    from ufl import MixedElement, VectorElement, TensorElement
    if isinstance(expr, (VectorElement, TensorElement)):
        element, (_, boundary) = get_constraints_metadata(expr)
        from dune.perftool.generation import get_global_context_value
        name = get_global_context_value("data").object_names.get(id(boundary), "zero")
        define_boundary_function(boundary, name)
        return name
    if isinstance(expr, MixedElement):
        children = tuple(name_boundary_function(el) for el in expr.sub_elements())
        name = '_'.join(children)
        define_compositegfs_parameterfunction(name, *children)
        return name

    # This is a scalar leaf of the ansatz function tree
    element, (_, boundary) = get_constraints_metadata(expr)

    from dune.perftool.generation import get_global_context_value
    name = get_global_context_value("data").object_names.get(id(boundary), "zero")

    define_boundary_function(boundary, name)
    return name


@symbol
def name_solution_function(expr):
    # Get expression representing solution
    from dune.perftool.generation import get_global_context_value
    solution_expression_name = get_option("exact_solution_expression")
    solution_expression = get_global_context_value("data").object_by_name[solution_expression_name]

    # Create function just like it is done for boundary_function
    name = "solution_function"
    define_boundary_function(solution_expression, name)

    return name


@preamble
def interpolate_vector(name):
    define_vector(name)
    element = _form.coefficients()[0].ufl_element()
    bf = name_boundary_function(element)
    gfs = name_gfs(element)
    return "Dune::PDELab::interpolate({}, {}, {});".format(bf,
                                                           gfs,
                                                           name,
                                                           )


@preamble
def interpolate_solution_expression(name):
    define_vector(name)
    element = _form.coefficients()[0].ufl_element()
    sol = name_solution_function(element)
    gfs = name_gfs(element)
    return "Dune::PDELab::interpolate({}, {}, {});".format(sol,
                                                           gfs,
                                                           name,
                                                           )


def maybe_interpolate_vector(name):
    element = _form.coefficients()[0].ufl_element()
    if has_constraints(element):
        interpolate_vector(name)
    else:
        define_vector(name)


@symbol
def name_vector():
    maybe_interpolate_vector("x")
    return "x"


@symbol
def name_vector_from_solution_expression():
    interpolate_solution_expression("solution")
    return "solution"


@preamble
def typedef_linearsolver(name):
    include_file("dune/pdelab/backend/istl.hh", filetag="driver")
    return "typedef Dune::PDELab::ISTLBackend_SEQ_UMFPack {};".format(name)


@symbol
def type_linearsolver():
    typedef_linearsolver("LinearSolver")
    return "LinearSolver"


@preamble
def define_linearsolver(name):
    lstype = type_linearsolver()
    return "{} {}(false);".format(lstype, name)


@symbol
def name_linearsolver():
    define_linearsolver("ls")
    return "ls"


@preamble
def define_reduction(name):
    ini = name_initree()
    return "double {} = {}.get<double>(\"reduction\", 1e-12);".format(name, ini)


@symbol
def name_reduction():
    define_reduction("reduction")
    return "reduction"


@preamble
def typedef_stationarylinearproblemsolver(name):
    include_file("dune/pdelab/stationary/linearproblem.hh", filetag="driver")
    gotype = type_gridoperator()
    lstype = type_linearsolver()
    xtype = type_vector()
    return "typedef Dune::PDELab::StationaryLinearProblemSolver<{}, {}, {}> {};".format(gotype, lstype, xtype, name)


@preamble
def typedef_stationarynonlinearproblemsolver(name):
    include_file("dune/pdelab/newton/newton.hh", filetag="driver")
    gotype = type_gridoperator()
    lstype = type_linearsolver()
    xtype = type_vector()
    return "typedef Dune::PDELab::Newton<{}, {}, {}> {};".format(gotype, lstype, xtype, name)


@symbol
def type_stationarylinearproblemsolver():
    typedef_stationarylinearproblemsolver("SLP")
    return "SLP"


@symbol
def type_stationarynonlinearproblemssolver():
    typedef_stationarynonlinearproblemsolver("SNP")
    return "SNP"


@preamble
def define_stationarylinearproblemsolver(name):
    slptype = type_stationarylinearproblemsolver()
    go = name_gridoperator()
    ls = name_linearsolver()
    x = name_vector()
    red = name_reduction()
    return "{} {}({}, {}, {}, {});".format(slptype, name, go, ls, x, red)


@preamble
def define_stationarynonlinearproblemsolver(name):
    snptype = type_stationarynonlinearproblemssolver()
    go = name_gridoperator()
    x = name_vector()
    ls = name_linearsolver()
    return "{}  {}({}, {}, {});".format(snptype, name, go, x, ls)

@symbol
def name_stationarylinearproblemsolver():
    define_stationarylinearproblemsolver("slp")
    return "slp"


@symbol
def name_stationarynonlinearproblemsolver():
    define_stationarynonlinearproblemsolver("snp")
    return "snp"


@preamble
def dune_solve():
    # Test if form is linear in ansatzfunction
    if is_linear(_form):
        if not get_option("matrix_free"):
            slp = name_stationarylinearproblemsolver()
            return "{}.apply();".format(slp)
        else:
            go = name_gridoperator()
            x = name_vector()
            include_file("dune/perftool/matrixfree.hh", filetag="drive")
            return "solveMatrixFree({},{});".format(go,x)
    else:
        snp = name_stationarynonlinearproblemsolver()
        return "{}.apply();".format(snp)


@preamble
def define_vtkfile(name):
    ini = name_initree()
    include_file("string", filetag="driver")
    return "std::string {} = {}.get<std::string>(\"wrapper.vtkcompare.name\", \"output\");".format(name, ini)


@symbol
def name_vtkfile():
    define_vtkfile("vtkfile")
    return "vtkfile"


@preamble
def compare_dofs():
    v = name_vector()
    solution = name_vector_from_solution_expression()
    return ["",
            "// Maximum norm of difference between dof vectors of the",
            "// numerical solution and the interpolation of the exact solution",
            "using Dune::PDELab::Backend::native;",
            "double maxerror(0.0);",
            "for (std::size_t i=0; i<native({}).size(); ++i)".format(v),
            "  if (std::abs(native({})[i]-native({})[i]) > maxerror)".format(v, solution),
            "    maxerror = std::abs(native({})[i]-native({})[i]);".format(v, solution),
            "std::cout << std::endl << \"Maxerror: \" << maxerror << std::endl;",
            "if (maxerror>{})".format(get_option("compare_dofs")),
            "  throw;"]


@preamble
def define_discrete_grid_function(gfs, vector_name, dgf_name):
    return ["typedef Dune::PDELab::DiscreteGridFunction<decltype({}),decltype({})> DGF;".format(gfs, vector_name),
            "DGF {}({},{});".format(dgf_name, gfs, vector_name)]


@symbol
def name_discrete_grid_function(gfs, vector_name):
    dgf_name = vector_name + "_dgf"
    define_discrete_grid_function(gfs, vector_name, dgf_name)
    return dgf_name


@preamble
def compare_L2_squared():
    element = _form.coefficients()[0].ufl_element()
    v = name_vector()
    gfs = name_gfs(element)
    vdgf = name_discrete_grid_function(gfs, v)
    solution_function = name_solution_function(element)

    include_file("dune/pdelab/gridfunctionspace/gridfunctionadapter.hh", filetag="driver")
    include_file("dune/pdelab/common/functionutilities.hh", filetag="driver")

    return ["",
            "// L2 error squared of difference between numerical",
            "// solution and the interpolation of exact solution",
            "typedef DifferenceSquaredAdapter<decltype({}),decltype({})> DifferenceSquared;".format(solution_function, vdgf),
            "DifferenceSquared differencesquared({},{});".format(solution_function, vdgf),
            "typename DifferenceSquared::Traits::RangeType l2errorsquared(0.0);",
            "Dune::PDELab::integrateGridFunction(differencesquared,l2errorsquared,10);",
            "std::cout << \"l2errorsquared: \" << l2errorsquared << std::endl;",
            "if (l2errorsquared>{})".format(get_option("compare_l2errorsquared")),
            "  throw;"]


@preamble
def print_residual():
    ini = name_initree()
    n_go = name_gridoperator()
    v = name_vector()
    t_v = type_vector()

    return ["if ({}.get<bool>(\"printresidual\", false)) {{".format(ini),
            "  using Dune::PDELab::Backend::native;",
            "  {} r({});".format(t_v, v),
            "  r=0.0;",
            "  {}.residual({}, r);".format(n_go, v),
            "  Dune::printvector(std::cout, native(r), \"residual vector\", \"row\");",
            "}"]


@preamble
def print_matrix():
    ini = name_initree()
    t_go = type_gridoperator()
    n_go = name_gridoperator()
    v = name_vector()
    t_v = type_vector()

    return ["if ({}.get<bool>(\"printmatrix\", false)) {{".format(ini),
            "  typedef typename {}::Traits::Jacobian M;".format(t_go),
            "  M m({});".format(n_go),
            "  {}.jacobian({},m);".format(n_go, v),
            "  using Dune::PDELab::Backend::native;",
            "  Dune::printmatrix(std::cout, native(m),\"global stiffness matrix\",\"row\",9,1);",
            "}"]


@preamble
def define_gfs_name(element, prefix=()):
    from ufl import MixedElement, VectorElement, TensorElement
    if isinstance(element, (VectorElement, TensorElement)):
        gfs = name_gfs(element)
        return "{}.name(\"data_{}\");".format(gfs, "_".join(str(p) for p in prefix))
    if isinstance(element, MixedElement):
        for (i, el) in enumerate(element.sub_elements()):
            define_gfs_name(el, prefix + (i,))
        return ""
    # This is a scalar leaf
    gfs = name_gfs(element)
    return "{}.name(\"data_{}\");".format(gfs, "_".join(str(p) for p in prefix))


@symbol
def type_predicate():
    include_file("dune/perftool/vtkpredicate.hh", filetag="driver")
    return "CuttingPredicate"


@preamble
def define_predicate(name):
    t = type_predicate()
    return "{} {};".format(t, name)


@symbol
def name_predicate():
    define_predicate("predicate")
    return "predicate"


@preamble
def vtkoutput():
    element = _form.coefficients()[0].ufl_element()
    define_gfs_name(element)
    include_file("dune/pdelab/gridfunctionspace/vtk.hh", filetag="driver")
    vtkwriter = name_vtkwriter()
    gfs = name_gfs(element)
    vec = name_vector()
    vtkfile = name_vtkfile()
    predicate = name_predicate()
    dune_solve()
    print_residual()
    print_matrix()

    if get_option("exact_solution_expression"):
        from ufl import MixedElement, VectorElement, TensorElement
        if isinstance(_form.coefficients()[0].ufl_element(),  (MixedElement, VectorElement, TensorElement)):
            raise NotImplementedError("Comparing to exact solution is olny implemented for scalar elements.")

        if get_option("compare_dofs"):
            compare_dofs()
        if get_option("compare_l2errorsquared"):
            compare_L2_squared()


    return ["Dune::PDELab::addSolutionToVTKWriter({}, {}, {}, Dune::PDELab::vtk::defaultNameScheme(), {});".format(vtkwriter, gfs, vec, predicate),
            "{}.write({}, Dune::VTK::ascii);".format(vtkwriter, vtkfile)]
