""" This module is loaded instead of ufl when executing .ufl files.
So, this module contains all our extensions and monkey patches to
UFL.
"""

import ufl

from ufl import *


# Monkey patch to the FiniteElementBase
# We want to subclass finite elements in order to allow meta data
# connected to constraints. This is an intermediate solution until
# we have a proper notion of (global) function spaces in UFL. As is,
# two finite elements with different meta data would not compare equally,
# but they should in order to allow hashing generation magic based on
# them (as done everywhere)
ufl.FiniteElementBase.__eq__ = lambda s, o: repr(s) == repr(o)


class TrialFunction(ufl.Coefficient):
    """ A coefficient that always takes the reserved index 0 """
    def __init__(self, element, count=None):
        if count and count is not 0:
            raise ValueError("The trial function must be the coefficient of index 0 in uflpdelab")
        ufl.Coefficient.__init__(self, element, count=0)


class ApplyCoefficient(ufl.Coefficient):
    """ A coefficient that always takes the reserved index 1 """
    def __init__(self, element, count=None):
        if count and count is not 1:
            raise ValueError("The jacobian apply coefficient must be of index 1 in uflpdelab")
        ufl.Coefficient.__init__(self, element, count=1)

class Coefficient(ufl.Coefficient):
    """ A coefficient that honors the reserved index 0. """
    def __init__(self, element, count=None):
        if count and count is 0:
            raise ValueError("The coefficient of index 0 is reserved for the trial function in uflpdelab")
        if count and count is 1:
            raise ValueError("The coefficient of index 1 is reserved for the jacobian apply vector in uflpdelab")
        if not count and ufl.Coefficient._globalcount < 2:
            count = 2
        ufl.Coefficient.__init__(self, element, count)


split = ufl.split_functions.split2


def Coefficients(element):
    return split(Coefficient(element))


def TestFunctions(element):
    return split(TestFunction(element))


def TrialFunctions(element):
    return split(TrialFunction(element))


class Expression(Coefficient):
    def __init__(self, expr, is_global=True, on_intersection=False, cell_type="triangle"):
        assert isinstance(expr, str)
        self.c_expr = expr
        self.is_global = is_global
        self.on_intersection = on_intersection

        # Avoid ufl complaining about not matching dimension/cells
        _dg0 = FiniteElement("DG", cell_type, 0)

        # Initialize a coefficient with a dummy finite element map.
        Coefficient.__init__(self, _dg0)

    # TODO the subdomain_data code needs a uflid, not idea how to get it here
    # The standard way through class decorator fails here...
    def ufl_id(self):
        return 0


class VectorExpression(Coefficient):
    def __init__(self, expr, is_global=True, on_intersection=False, cell_type="triangle"):
        assert isinstance(expr, str)
        self.c_expr = expr
        self.is_global = is_global
        self.on_intersection = on_intersection

        # Avoid ufl complaining about not matching dimension/cells
        _dg0vec = VectorElement("DG", cell_type, 0)

        # Initialize a coefficient with a dummy finite element map.
        Coefficient.__init__(self, _dg0vec)

    # TODO the subdomain_data code needs a uflid, not idea how to get it here
    # The standard way through class decorator fails here...
    def ufl_id(self):
        return 0


class FiniteElement(ufl.FiniteElement):
    def __init__(self, *args, **kwargs):
        if ('dirichlet_constraints' in kwargs) or ('dirichlet_expression' in kwargs):
            # Get dirichlet_constraints and convert it to Expression if necessary!
            self.dirichlet_constraints = kwargs.pop('dirichlet_constraints', 'return true;')
            if isinstance(self.dirichlet_constraints, str):
                self.dirichlet_constraints = Expression(self.dirichlet_constraints)
            assert isinstance(self.dirichlet_constraints, Expression)

            # Get dirichlet_constraints and convert it to Expression if necessary!
            self.dirichlet_expression = kwargs.pop('dirichlet_expression', 'return 0.0;')
            if isinstance(self.dirichlet_expression, str):
                self.dirichlet_expression = Expression(self.dirichlet_expression)
            assert isinstance(self.dirichlet_expression, Expression)

        # Initialize the original finite element from ufl
        ufl.FiniteElement.__init__(self, *args, **kwargs)


class VectorElement(ufl.VectorElement):
    def __init__(self, *args, **kwargs):
        if ('dirichlet_constraints' in kwargs) or ('dirichlet_expression' in kwargs):
            # Get dirichlet_constraints and convert it to Expression if necessary!
            self.dirichlet_constraints = kwargs.pop('dirichlet_constraints', 'return true;')
            if isinstance(self.dirichlet_constraints, str):
                self.dirichlet_constraints = Expression(self.dirichlet_constraints)
            assert isinstance(self.dirichlet_constraints, Expression)

            # Get dirichlet_constraints and convert it to Expression if necessary!
            self.dirichlet_expression = kwargs.pop('dirichlet_expression', 'return 0.0;')
            if isinstance(self.dirichlet_expression, str):
                self.dirichlet_expression = Expression(self.dirichlet_expression)
            assert isinstance(self.dirichlet_expression, Expression)

        # Initialize the original finite element from ufl
        ufl.VectorElement.__init__(self, *args, **kwargs)
